@@ -1,39 +1,97 @@
 use tokio::task::JoinHandle;
 use std::hash::Hash;
 use futures::Future;
+use thiserror::Error;
 use crate::internal_cache::{CacheAction, InternalCacheStore, CacheMessage};
 use crate::backing::{CacheBacking, HashMapBacking};
+use std::fmt::Debug;
 
-#[derive(Debug, Clone)]
-pub struct CacheLoadingError {
-    pub reason_phrase: String,
-    // todo: nested errors
+#[derive(Error, Debug)]
+pub enum CacheLoadingError<E: Debug> {
+    #[error(transparent)]
+    CommunicationError(CacheCommunicationError),
+    #[error("No data found")]
+    NoData(), // todo better handling here? eventually return loadingerror if possible
+    #[error("An error occurred when loading the entity from the loader function")]
+    LoadingError(E)
+}
+
+#[derive(Error, Debug)]
+pub enum CacheCommunicationError {
+    #[error("An error occurred when trying to submit the cache request")]
+    TokioMpscSendError(),
+    #[error("An error occurred when trying to join the result future")]
+    FutureJoinError(#[from] tokio::task::JoinError),
+    #[error("An error occurred when waiting for the broadcaster response")]
+    TokioBroadcastRecvError(#[from] tokio::sync::broadcast::error::RecvError),
+    #[error("An error occurred when receiving the response")]
+    TokioOneshotRecvError(#[from] tokio::sync::oneshot::error::RecvError),
+    #[error("Lookups are looping, internal error")]
+    LookupLoop(),
+}
+
+impl<E: Debug> CacheLoadingError<E> {
+
+    pub fn as_loading_error(&self) -> Option<&E> {
+        match self {
+            CacheLoadingError::LoadingError(error) => Some(error),
+            _ => None
+        }
+    }
+
+    pub fn into_loading_error(self) -> Option<E> {
+        match self {
+            CacheLoadingError::LoadingError(error) => Some(error),
+            _ => None
+        }
+    }
+
+    pub fn as_communication_error(&self) -> Option<&CacheCommunicationError> {
+        match self {
+            CacheLoadingError::CommunicationError(error) => Some(error),
+            _ => None
+        }
+    }
+
+    pub fn into_communication_error(self) -> Option<CacheCommunicationError> {
+        match self {
+            CacheLoadingError::CommunicationError(error) => Some(error),
+            _ => None
+        }
+    }
+}
+
+#[derive(Clone)]
+pub struct ResultMeta<V> {
+    pub result: V,
+    pub cached: bool,
 }
 
 #[derive(Debug, Clone)]
-pub enum CacheEntry<V> {
+pub enum CacheEntry<V, E: Debug> {
     Loaded(V),
-    Loading(tokio::sync::broadcast::Sender<Option<V>>),
+    Loading(tokio::sync::broadcast::Sender<Result<V, E>>),
 }
 
 #[derive(Debug)]
-pub enum CacheResult<V> {
+pub enum CacheResult<V, E: Debug> {
     Found(V),
-    Loading(JoinHandle<Result<V, CacheLoadingError>>),
+    Loading(JoinHandle<Result<V, CacheLoadingError<E>>>),
     None,
 }
 
 pub type CacheHandle = JoinHandle<()>;
 
 #[derive(Debug, Clone)]
-pub struct LoadingCache<K, V> {
-    tx: tokio::sync::mpsc::Sender<CacheMessage<K, V>>
+pub struct LoadingCache<K, V, E: Debug> {
+    tx: tokio::sync::mpsc::Sender<CacheMessage<K, V, E>>
 }
 
 impl<
     K: Eq + Hash + Clone + Send + 'static,
     V: Clone + Sized + Send + 'static,
-> LoadingCache<K, V> {
+    E: Clone + Sized + Send + Debug + 'static,
+> LoadingCache<K, V, E> {
     /// Creates a new instance of a LoadingCache with the default `HashMapBacking`
     ///
     /// # Arguments
@@ -61,7 +119,7 @@ impl<
     ///     let (cache, _) = LoadingCache::new(move |key: String| {
     ///         let db_clone = static_db.clone();
     ///         async move {
-    ///             db_clone.get(&key).cloned()
+    ///             db_clone.get(&key).cloned().ok_or(1)
     ///         }
     ///     });
     ///
@@ -70,8 +128,8 @@ impl<
     ///     assert_eq!(result, 32);
     /// }
     /// ```
-    pub fn new<T, F>(loader: T) -> (LoadingCache<K, V>, CacheHandle)
-        where F: Future<Output=Option<V>> + Sized + Send + 'static,
+    pub fn new<T, F>(loader: T) -> (LoadingCache<K, V, E>, CacheHandle)
+        where F: Future<Output=Result<V, E>> + Sized + Send + 'static,
               T: Fn(K) -> F + Send + 'static {
         LoadingCache::with_backing(HashMapBacking::new(), loader)
     }
@@ -107,7 +165,7 @@ impl<
     ///         move |key: String| {
     ///             let db_clone = static_db.clone();
     ///             async move {
-    ///                 db_clone.get(&key).cloned()
+    ///                 db_clone.get(&key).cloned().ok_or(1)
     ///             }
     ///         }
     ///     );
@@ -117,10 +175,10 @@ impl<
     ///     assert_eq!(result, 32);
     /// }
     /// ```
-    pub fn with_backing<T, F, B>(backing: B, loader: T) -> (LoadingCache<K, V>, CacheHandle)
-        where F: Future<Output=Option<V>> + Sized + Send + 'static,
+    pub fn with_backing<T, F, B>(backing: B, loader: T) -> (LoadingCache<K, V, E>, CacheHandle)
+        where F: Future<Output=Result<V, E>> + Sized + Send + 'static,
               T: Fn(K) -> F + Send + 'static,
-              B: CacheBacking<K, CacheEntry<V>> + Send + 'static {
+              B: CacheBacking<K, CacheEntry<V, E>> + Send + 'static {
         let (tx, rx) = tokio::sync::mpsc::channel(128);
         let store = InternalCacheStore::new(backing, tx.clone(), loader);
         let handle = store.run(rx);
@@ -140,7 +198,13 @@ impl<
     /// Returns a Result with:
     /// Ok - Value of type V
     /// Err - Error of type CacheLoadingError
-    pub async fn get(&self, key: K) -> Result<V, CacheLoadingError> {
+    pub async fn get(&self, key: K) -> Result<V, CacheLoadingError<E>> {
+        self.send_cache_action(CacheAction::Get(key)).await
+            .map(|opt_result| opt_result.expect("Get should always return either V or CacheLoadingError"))
+            .map(|meta| meta.result)
+    }
+
+    pub async fn get_with_meta(&self, key: K) -> Result<ResultMeta<V>, CacheLoadingError<E>> {
         self.send_cache_action(CacheAction::Get(key)).await
             .map(|opt_result| opt_result.expect("Get should always return either V or CacheLoadingError"))
     }
@@ -159,8 +223,9 @@ impl<
     /// Ok - Previous value of type V wrapped in an Option depending whether there was a previous
     ///      value
     /// Err - Error of type CacheLoadingError
-    pub async fn set(&self, key: K, value: V) -> Result<Option<V>, CacheLoadingError> {
+    pub async fn set(&self, key: K, value: V) -> Result<Option<V>, CacheLoadingError<E>> {
         self.send_cache_action(CacheAction::Set(key, value)).await
+            .map(|opt_meta| opt_meta.map(|meta| meta.result))
     }
 
     /// Loads the value for the specified key from the cache and returns None if not present
@@ -174,8 +239,9 @@ impl<
     /// Returns a Result with:
     /// Ok - Value of type Option<V>
     /// Err - Error of type CacheLoadingError
-    pub async fn get_if_present(&self, key: K) -> Result<Option<V>, CacheLoadingError> {
+    pub async fn get_if_present(&self, key: K) -> Result<Option<V>, CacheLoadingError<E>> {
         self.send_cache_action(CacheAction::GetIfPresent(key)).await
+            .map(|opt_meta| opt_meta.map(|meta| meta.result))
     }
 
     /// Checks whether a specific value is mapped for the given key
@@ -189,7 +255,7 @@ impl<
     /// Returns a Result with:
     /// Ok - bool
     /// Err - Error of type CacheLoadingError
-    pub async fn exists(&self, key: K) -> Result<bool, CacheLoadingError> {
+    pub async fn exists(&self, key: K) -> Result<bool, CacheLoadingError<E>> {
         self.get_if_present(key).await
             .map(|result| result.is_some())
     }
@@ -206,8 +272,9 @@ impl<
     /// Returns a Result with:
     /// Ok - Value of type Option<V>
     /// Err - Error of type CacheLoadingError
-    pub async fn remove(&self, key: K) -> Result<Option<V>, CacheLoadingError> {
+    pub async fn remove(&self, key: K) -> Result<Option<V>, CacheLoadingError<E>> {
         self.send_cache_action(CacheAction::Remove(key)).await
+            .map(|opt_meta| opt_meta.map(|meta| meta.result))
     }
 
     /// Updates a key on the cache with the given update function and returns the previous value
@@ -229,19 +296,21 @@ impl<
     /// Returns a Result with:
     /// Ok - Value of type V which is the previously mapped value
     /// Err - Error of type CacheLoadingError
-    pub async fn update<U>(&self, key: K, update_fn: U) -> Result<V, CacheLoadingError>
+    pub async fn update<U>(&self, key: K, update_fn: U) -> Result<V, CacheLoadingError<E>>
         where U: FnOnce(V) -> V + Send + 'static {
         self.send_cache_action(CacheAction::Update(key, Box::new(update_fn))).await
             .map(|opt_result| opt_result.expect("Get should always return either V or CacheLoadingError"))
+            .map(|meta| meta.result)
     }
 
-    pub async fn update_mut<U>(&self, key: K, update_fn: U) -> Result<V, CacheLoadingError>
+    pub async fn update_mut<U>(&self, key: K, update_fn: U) -> Result<V, CacheLoadingError<E>>
         where U: FnMut(&mut V) -> () + Send + 'static {
         self.send_cache_action(CacheAction::UpdateMut(key, Box::new(update_fn))).await
             .map(|opt_result| opt_result.expect("Get should always return either V or CacheLoadingError"))
+            .map(|meta| meta.result)
     }
 
-    async fn send_cache_action(&self, action: CacheAction<K, V>) -> Result<Option<V>, CacheLoadingError> {
+    async fn send_cache_action(&self, action: CacheAction<K, V>) -> Result<Option<ResultMeta<V>>, CacheLoadingError<E>> {
         let (tx, rx) = tokio::sync::oneshot::channel();
         match self.tx.send(CacheMessage {
             action,
@@ -251,28 +320,36 @@ impl<
                 match rx.await {
                     Ok(result) => {
                         match result {
-                            CacheResult::Found(value) => { Ok(Some(value)) }
+                            CacheResult::Found(value) => {
+                                Ok(Some(ResultMeta {
+                                    result: value,
+                                    cached: true,
+                                }))
+                            }
                             CacheResult::Loading(handle) => {
                                 match handle.await {
                                     Ok(load_result) => {
-                                        load_result.map(|v| Some(v))
+                                        load_result.map(|v| Some(ResultMeta {
+                                            result: v,
+                                            cached: false,
+                                        }))
                                     }
-                                    Err(_) => {
-                                        Err(CacheLoadingError { reason_phrase: "Error when trying to join loader future".to_owned() })
+                                    Err(err) => {
+                                        Err(CacheLoadingError::CommunicationError(CacheCommunicationError::FutureJoinError(err)))
                                     }
                                 }
                             }
                             CacheResult::None => { Ok(None) }
                         }
                     }
-                    Err(_) => {
-                        Err(CacheLoadingError { reason_phrase: "Error when receiving cache response".to_owned() })
+                    Err(err) => {
+                        Err(CacheLoadingError::CommunicationError(CacheCommunicationError::TokioOneshotRecvError(err)))
                     }
                 }
             }
             Err(_) => {
-                Err(CacheLoadingError { reason_phrase: "Error when trying to submit cache request".to_owned() })
+                Err(CacheLoadingError::CommunicationError(CacheCommunicationError::TokioMpscSendError()))
             }
         }
     }
-}
\ No newline at end of file
+}
