@@ -1,8 +1,9 @@
 use std::hash::Hash;
 use futures::Future;
 use tokio::task::JoinHandle;
-use crate::cache_api::{CacheResult, CacheLoadingError, CacheEntry};
+use crate::cache_api::{CacheResult, CacheLoadingError, CacheEntry, CacheCommunicationError};
 use crate::backing::CacheBacking;
+use std::fmt::Debug;
 
 pub(crate) enum CacheAction<K, V> {
     GetIfPresent(K),
@@ -16,13 +17,13 @@ pub(crate) enum CacheAction<K, V> {
     Unblock(K),
 }
 
-pub(crate) struct CacheMessage<K, V> {
+pub(crate) struct CacheMessage<K, V, E: Debug> {
     pub(crate) action: CacheAction<K, V>,
-    pub(crate) response: tokio::sync::oneshot::Sender<CacheResult<V>>,
+    pub(crate) response: tokio::sync::oneshot::Sender<CacheResult<V, E>>,
 }
 
-pub(crate) struct InternalCacheStore<K, V, T, B> {
-    tx: tokio::sync::mpsc::Sender<CacheMessage<K, V>>,
+pub(crate) struct InternalCacheStore<K, V, T, B, E: Debug> {
+    tx: tokio::sync::mpsc::Sender<CacheMessage<K, V, E>>,
     data: B,
     loader: T,
 }
@@ -30,14 +31,15 @@ pub(crate) struct InternalCacheStore<K, V, T, B> {
 impl<
     K: Eq + Hash + Clone + Send + 'static,
     V: Clone + Sized + Send + 'static,
-    F: Future<Output=Option<V>> + Sized + Send + 'static,
+    E: Clone + Sized + Send + Debug + 'static,
+    F: Future<Output=Result<V, E>> + Sized + Send + 'static,
     T: Fn(K) -> F + Send + 'static,
-    B: CacheBacking<K, CacheEntry<V>> + Send + 'static
-> InternalCacheStore<K, V, T, B>
+    B: CacheBacking<K, CacheEntry<V, E>> + Send + 'static
+> InternalCacheStore<K, V, T, B, E>
 {
     pub fn new(
         backing: B,
-        tx: tokio::sync::mpsc::Sender<CacheMessage<K, V>>,
+        tx: tokio::sync::mpsc::Sender<CacheMessage<K, V, E>>,
         loader: T,
     ) -> Self {
         Self {
@@ -47,7 +49,7 @@ impl<
         }
     }
 
-    pub(crate) fn run(mut self, mut rx: tokio::sync::mpsc::Receiver<CacheMessage<K, V>>) -> JoinHandle<()> {
+    pub(crate) fn run(mut self, mut rx: tokio::sync::mpsc::Receiver<CacheMessage<K, V, E>>) -> JoinHandle<()> {
         tokio::spawn(async move {
             loop {
                 if let Some(message) = rx.recv().await {
@@ -72,14 +74,17 @@ impl<
 
     fn unblock(&mut self, key: K) {
         if let Some(entry) = self.data.get(&key) {
-            if let CacheEntry::Loading(waiter) = entry {
-                waiter.send(None).ok();
-                self.data.remove(&key);
+            if let CacheEntry::Loading(_) = entry {
+                if let Some(entry) = self.data.remove(&key) {
+                    if let CacheEntry::Loading(waiter) = entry {
+                        std::mem::drop(waiter) // dropping the sender closes the channel
+                    }
+                }
             }
         }
     }
 
-    fn remove(&mut self, key: K) -> CacheResult<V> {
+    fn remove(&mut self, key: K) -> CacheResult<V, E> {
         if let Some(entry) = self.data.remove(&key) {
             match entry {
                 CacheEntry::Loaded(data) => CacheResult::Found(data),
@@ -90,7 +95,7 @@ impl<
         }
     }
 
-    fn update_mut(&mut self, key: K, mut update_mut_fn: Box<dyn FnMut(&mut V) -> () + Send + 'static>) -> CacheResult<V> {
+    fn update_mut(&mut self, key: K, mut update_mut_fn: Box<dyn FnMut(&mut V) -> () + Send + 'static>) -> CacheResult<V, E> {
         match self.data.get_mut(&key) {
             Some(entry) => {
                 match entry {
@@ -106,11 +111,11 @@ impl<
                             let (response_tx, response_rx) = tokio::sync::oneshot::channel();
                             cache_tx.send(CacheMessage {
                                 action: CacheAction::UpdateMut(key, update_mut_fn),
-                                response: response_tx
+                                response: response_tx,
                             }).await.ok();
                             match response_rx.await.unwrap() {
                                 CacheResult::Found(data) => Ok(data),
-                                _ => Err(CacheLoadingError { reason_phrase: "No data found".to_owned() })
+                                _ => Err(CacheLoadingError::NoData())
                             }
                         }))
                     }
@@ -126,11 +131,11 @@ impl<
                             let (response_tx, response_rx) = tokio::sync::oneshot::channel();
                             cache_tx.send(CacheMessage {
                                 action: CacheAction::UpdateMut(key, update_mut_fn),
-                                response: response_tx
+                                response: response_tx,
                             }).await.ok();
                             match response_rx.await.unwrap() {
                                 CacheResult::Found(data) => Ok(data),
-                                _ => Err(CacheLoadingError { reason_phrase: "No data found".to_owned() })
+                                _ => Err(CacheLoadingError::NoData())
                             }
                         }))
                     }
@@ -140,7 +145,7 @@ impl<
         }
     }
 
-    fn update(&mut self, key: K, update_fn: Box<dyn FnOnce(V) -> V + Send + 'static>) -> CacheResult<V> {
+    fn update(&mut self, key: K, update_fn: Box<dyn FnOnce(V) -> V + Send + 'static>) -> CacheResult<V, E> {
         match self.get(key.clone()) {
             CacheResult::Found(data) => {
                 let updated_data = update_fn(data);
@@ -163,17 +168,11 @@ impl<
                         Ok(result) => {
                             match result {
                                 CacheResult::Found(data) => Ok(data),
-                                CacheResult::Loading(_) => Err(CacheLoadingError {
-                                    reason_phrase: "2nd lookup is not available".to_string()
-                                }),
-                                CacheResult::None => Err(CacheLoadingError {
-                                    reason_phrase: "2nd lookup is not available".to_string()
-                                })
+                                CacheResult::Loading(_) => Err(CacheLoadingError::CommunicationError(CacheCommunicationError::LookupLoop())),
+                                CacheResult::None => Err(CacheLoadingError::CommunicationError(CacheCommunicationError::LookupLoop()))
                             }
                         }
-                        Err(_) => Err(CacheLoadingError {
-                            reason_phrase: "Error when receiving response".to_string()
-                        }),
+                        Err(err) => Err(CacheLoadingError::CommunicationError(CacheCommunicationError::TokioOneshotRecvError(err))),
                     }
                 }))
             }
@@ -181,7 +180,7 @@ impl<
         }
     }
 
-    fn set(&mut self, key: K, value: V, loading_result: bool) -> CacheResult<V> {
+    fn set(&mut self, key: K, value: V, loading_result: bool) -> CacheResult<V, E> {
         let opt_entry = self.data.get(&key);
         if loading_result {
             if opt_entry.is_none() {
@@ -203,7 +202,7 @@ impl<
             .unwrap_or(CacheResult::None)
     }
 
-    fn get_if_present(&mut self, key: K) -> CacheResult<V> {
+    fn get_if_present(&mut self, key: K) -> CacheResult<V, E> {
         if let Some(entry) = self.data.get(&key) {
             match entry {
                 CacheEntry::Loaded(data) => CacheResult::Found(data.clone()),
@@ -214,7 +213,7 @@ impl<
         }
     }
 
-    fn get(&mut self, key: K) -> CacheResult<V> {
+    fn get(&mut self, key: K) -> CacheResult<V, E> {
         if let Some(entry) = self.data.get(&key) {
             match entry {
                 CacheEntry::Loaded(value) => {
@@ -223,14 +222,18 @@ impl<
                 CacheEntry::Loading(waiter) => {
                     let waiter = waiter.clone();
                     CacheResult::Loading(tokio::spawn(async move {
-                        if let Ok(result) = waiter.subscribe().recv().await {
-                            if let Some(data) = result {
-                                Ok(data)
-                            } else {
-                                Err(CacheLoadingError { reason_phrase: "Loader function returned None".to_owned() })
+                        match waiter.subscribe().recv().await {
+                            Ok(result) => {
+                                match result {
+                                    Ok(data) => {
+                                        Ok(data)
+                                    }
+                                    Err(loading_error) => {
+                                        Err(CacheLoadingError::LoadingError(loading_error))
+                                    }
+                                }
                             }
-                        } else {
-                            Err(CacheLoadingError { reason_phrase: "Waiter broadcast channel error".to_owned() })
+                            Err(err) => Err(CacheLoadingError::CommunicationError(CacheCommunicationError::TokioBroadcastRecvError(err)))
                         }
                     }))
                 }
@@ -242,25 +245,28 @@ impl<
             let loader = (self.loader)(key.clone());
             let inner_key = key.clone();
             let join_handle = tokio::spawn(async move {
-                if let Some(value) = loader.await {
-                    inner_tx.send(Some(value.clone())).ok();
-                    let (tx, rx) = tokio::sync::oneshot::channel();
-                    let send_value = value.clone();
-                    cache_tx.send(CacheMessage {
-                        action: CacheAction::SetAndUnblock(inner_key, send_value),
-                        response: tx,
-                    }).await.ok();
-                    rx.await.ok(); // await cache confirmation
-                    Ok(value)
-                } else {
-                    inner_tx.send(None).ok();
-                    let (tx, rx) = tokio::sync::oneshot::channel();
-                    cache_tx.send(CacheMessage {
-                        action: CacheAction::Unblock(inner_key),
-                        response: tx,
-                    }).await.ok();
-                    rx.await.ok(); // await cache confirmation
-                    Err(CacheLoadingError { reason_phrase: "Loader function returned None".to_owned() })
+                match loader.await {
+                    Ok(value) => {
+                        inner_tx.send(Ok(value.clone())).ok();
+                        let (tx, rx) = tokio::sync::oneshot::channel();
+                        let send_value = value.clone();
+                        cache_tx.send(CacheMessage {
+                            action: CacheAction::SetAndUnblock(inner_key, send_value),
+                            response: tx,
+                        }).await.ok();
+                        rx.await.ok(); // await cache confirmation
+                        Ok(value)
+                    }
+                    Err(loading_error) => {
+                        inner_tx.send(Err(loading_error.clone())).ok();
+                        let (tx, rx) = tokio::sync::oneshot::channel();
+                        cache_tx.send(CacheMessage {
+                            action: CacheAction::Unblock(inner_key),
+                            response: tx,
+                        }).await.ok();
+                        rx.await.ok(); // await cache confirmation
+                        Err(CacheLoadingError::LoadingError(loading_error))
+                    }
                 }
             });
             self.data.set(key, CacheEntry::Loading(tx));
